@@ -1,6 +1,6 @@
 /* -*- c-basic-offset: 2 -*- */
 /*
-  Copyright(C) 2012-2016 Brazil
+  Copyright(C) 2012-2018 Brazil
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -19,7 +19,7 @@
 #pragma once
 
 #include <groonga/plugin.h>
-#include <groonga/token.h>
+#include <groonga/tokenizer_query_deprecated.h>
 
 #ifdef __cplusplus
 extern "C" {
@@ -69,25 +69,6 @@ GRN_PLUGIN_EXPORT grn_bool grn_tokenizer_have_tokenized_delimiter(grn_ctx *ctx,
                                                                   unsigned int str_length,
                                                                   grn_encoding encoding);
 
-/*
-  grn_tokenizer_query is a structure for storing a query. See the following
-  functions.
- */
-typedef struct _grn_tokenizer_query grn_tokenizer_query;
-
-struct _grn_tokenizer_query {
-  grn_obj *normalized_query;
-  char *query_buf;
-  const char *ptr;
-  unsigned int length;
-  grn_encoding encoding;
-  unsigned int flags;
-  grn_bool have_tokenized_delimiter;
-  /* Deprecated since 4.0.8. Use tokenize_mode instead. */
-  grn_token_mode token_mode;
-  grn_tokenize_mode tokenize_mode;
-};
-
 /*
   grn_tokenizer_query_open() parses `args' and returns a new object of
   grn_tokenizer_query. The new object stores information of the query.
@@ -122,6 +103,39 @@ GRN_PLUGIN_EXPORT void grn_tokenizer_query_close(grn_ctx *ctx, grn_tokenizer_que
  */
 void grn_tokenizer_query_destroy(grn_ctx *ctx, grn_tokenizer_query *query);
 
+GRN_PLUGIN_EXPORT grn_rc
+grn_tokenizer_query_set_normalize_flags(grn_ctx *ctx,
+                                        grn_tokenizer_query *query,
+                                        unsigned int flags);
+GRN_PLUGIN_EXPORT unsigned int
+grn_tokenizer_query_get_normalize_flags(grn_ctx *ctx,
+                                        grn_tokenizer_query *query);
+
+GRN_PLUGIN_EXPORT grn_obj *
+grn_tokenizer_query_get_normalized_string(grn_ctx *ctx,
+                                          grn_tokenizer_query *query);
+
+GRN_PLUGIN_EXPORT const char *
+grn_tokenizer_query_get_raw_string(grn_ctx *ctx,
+                                   grn_tokenizer_query *query,
+                                   size_t *length);
+
+GRN_PLUGIN_EXPORT grn_encoding
+grn_tokenizer_query_get_encoding(grn_ctx *ctx, grn_tokenizer_query *query);
+
+GRN_PLUGIN_EXPORT unsigned int
+grn_tokenizer_query_get_flags(grn_ctx *ctx, grn_tokenizer_query *query);
+
+GRN_PLUGIN_EXPORT grn_bool
+grn_tokenizer_query_have_tokenized_delimiter(grn_ctx *ctx,
+                                             grn_tokenizer_query *query);
+
+GRN_PLUGIN_EXPORT grn_tokenize_mode
+grn_tokenizer_query_get_mode(grn_ctx *ctx, grn_tokenizer_query *query);
+
+GRN_PLUGIN_EXPORT grn_obj *
+grn_tokenizer_query_get_lexicon(grn_ctx *ctx, grn_tokenizer_query *query);
+
 /*
   grn_tokenizer_token is needed to return tokens. A grn_tokenizer_token object
   stores a token to be returned and it must be maintained until a request for
@@ -249,11 +263,43 @@ GRN_PLUGIN_EXPORT const char *grn_tokenizer_tokenized_delimiter_next(grn_ctx *ct
   GRN_SUCCESS on success, an error code on failure. See "groonga.h" for more
   details of grn_proc_func and grn_user_data, that is used as an argument of
   grn_proc_func.
+
+  Deprecated since 8.0.2. Use grn_tokenizer_create() and
+  grn_tokenizer_set_*_func().
  */
-GRN_PLUGIN_EXPORT grn_rc grn_tokenizer_register(grn_ctx *ctx, const char *plugin_name_ptr,
-                                                unsigned int plugin_name_length,
-                                                grn_proc_func *init, grn_proc_func *next,
-                                                grn_proc_func *fin);
+GRN_PLUGIN_EXPORT grn_rc
+grn_tokenizer_register(grn_ctx *ctx, const char *plugin_name_ptr,
+                       unsigned int plugin_name_length,
+                       grn_proc_func *init, grn_proc_func *next,
+                       grn_proc_func *fin);
+
+GRN_PLUGIN_EXPORT grn_obj *
+grn_tokenizer_create(grn_ctx *ctx,
+                     const char *name,
+                     int name_length);
+
+typedef void *grn_tokenizer_init_func(grn_ctx *ctx,
+                                      grn_tokenizer_query *query);
+typedef void grn_tokenizer_next_func(grn_ctx *ctx,
+                                     grn_tokenizer_query *query,
+                                     grn_token *token,
+                                     void *user_data);
+typedef void grn_tokenizer_fin_func(grn_ctx *ctx,
+                                    void *user_data);
+
+
+GRN_PLUGIN_EXPORT grn_rc
+grn_tokenizer_set_init_func(grn_ctx *ctx,
+                            grn_obj *tokenizer,
+                            grn_tokenizer_init_func *init);
+GRN_PLUGIN_EXPORT grn_rc
+grn_tokenizer_set_next_func(grn_ctx *ctx,
+                            grn_obj *tokenizer,
+                            grn_tokenizer_next_func *next);
+GRN_PLUGIN_EXPORT grn_rc
+grn_tokenizer_set_fin_func(grn_ctx *ctx,
+                           grn_obj *tokenizer,
+                           grn_tokenizer_fin_func *fin);
 
 #ifdef __cplusplus
 }  /* extern "C" */
