@@ -1,6 +1,7 @@
 /* -*- c-basic-offset: 2 -*- */
 /*
-  Copyright(C) 2009-2017 Brazil
+  Copyright(C) 2009-2018 Brazil
+  Copyright(C) 2018 Kouhei Sutou <kou@clear-code.com>
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -128,6 +129,7 @@ typedef int64_t off64_t;
 
 typedef SOCKET grn_sock;
 # define grn_sock_close(sock) closesocket(sock)
+# define GRN_INVALID_SOCKET INVALID_SOCKET
 
 # define CALLBACK __stdcall
 
@@ -164,6 +166,7 @@ typedef unsigned char uint_least8_t;
 # endif /* UINT_LEAST8_MAX */
 typedef int grn_sock;
 # define grn_sock_close(sock) close(sock)
+# define GRN_INVALID_SOCKET -1
 # define CALLBACK
 
 #endif /* WIN32 */
@@ -226,38 +229,18 @@ typedef pthread_mutex_t grn_mutex;
 # define MUTEX_LOCK_CHECK(m) (MUTEX_LOCK(m) == 0)
 # define MUTEX_UNLOCK(m)     pthread_mutex_unlock(&(m))
 # define MUTEX_FIN(m)        pthread_mutex_destroy(&(m))
-# ifdef HAVE_PTHREAD_MUTEXATTR_SETPSHARED
-#  define MUTEX_INIT_SHARED(m) do {\
-  pthread_mutexattr_t mutexattr;\
-  pthread_mutexattr_init(&mutexattr);\
-  pthread_mutexattr_setpshared(&mutexattr, PTHREAD_PROCESS_SHARED);\
-  pthread_mutex_init(&(m), &mutexattr);\
-} while (0)
-# else
-#  define MUTEX_INIT_SHARED MUTEX_INIT
-# endif /* HAVE_PTHREAD_MUTEXATTR_SETPSHARED */
 
 typedef pthread_mutex_t grn_critical_section;
 # define CRITICAL_SECTION_INIT(cs)  pthread_mutex_init(&(cs), NULL)
 # define CRITICAL_SECTION_ENTER(cs) pthread_mutex_lock(&(cs))
 # define CRITICAL_SECTION_LEAVE(cs) pthread_mutex_unlock(&(cs))
-# define CRITICAL_SECTION_FIN(cs)
+# define CRITICAL_SECTION_FIN(cs)   pthread_mutex_destroy(&(cs))
 
 typedef pthread_cond_t grn_cond;
 # define COND_INIT(c)   pthread_cond_init(&(c), NULL)
 # define COND_SIGNAL(c) pthread_cond_signal(&(c))
-# define COND_WAIT(c,m) pthread_cond_wait(&(c), &(m))
+# define COND_WAIT(c,cs) pthread_cond_wait(&(c), &(cs))
 # define COND_BROADCAST(c) pthread_cond_broadcast(&(c))
-# ifdef HAVE_PTHREAD_CONDATTR_SETPSHARED
-#  define COND_INIT_SHARED(c) do {\
-  pthread_condattr_t condattr;\
-  pthread_condattr_init(&condattr);\
-  pthread_condattr_setpshared(&condattr, PTHREAD_PROCESS_SHARED);\
-  pthread_cond_init(&(c), &condattr);\
-} while (0)
-# else
-#  define COND_INIT_SHARED COND_INIT
-# endif /* HAVE_PTHREAD_CONDATTR_SETPSHARE */
 # define COND_FIN(c)    pthread_cond_destroy(&(c))
 
 typedef pthread_key_t grn_thread_key;
@@ -327,77 +310,12 @@ typedef CRITICAL_SECTION grn_critical_section;
 #  define CRITICAL_SECTION_LEAVE(cs) LeaveCriticalSection(&(cs))
 #  define CRITICAL_SECTION_FIN(cs)   DeleteCriticalSection(&(cs))
 
-typedef struct
-{
-  int waiters_count_;
-  HANDLE waiters_count_lock_;
-  HANDLE sema_;
-  HANDLE waiters_done_;
-  size_t was_broadcast_;
-} grn_cond;
-
-#  define COND_INIT(c) do { \
-  (c).waiters_count_ = 0; \
-  (c).sema_ = CreateSemaphore(NULL, 0, 0x7fffffff, NULL); \
-  MUTEX_INIT((c).waiters_count_lock_); \
-  (c).waiters_done_ = CreateEvent(NULL, FALSE, FALSE, NULL); \
-} while (0)
-
-#  define COND_SIGNAL(c) do { \
-  MUTEX_LOCK((c).waiters_count_lock_); \
-  { \
-    int have_waiters = (c).waiters_count_ > 0; \
-    MUTEX_UNLOCK((c).waiters_count_lock_); \
-    if (have_waiters) { \
-      ReleaseSemaphore((c).sema_, 1, 0); \
-    } \
-  } \
-} while (0)
-
-#  define COND_BROADCAST(c) do { \
-  MUTEX_LOCK((c).waiters_count_lock_); \
-  { \
-    int have_waiters = (c).waiters_count_ > 0; \
-    if ((c).waiters_count_ > 0) { \
-      (c).was_broadcast_ = 1; \
-      have_waiters = 1; \
-    } \
-    if (have_waiters) { \
-      ReleaseSemaphore((c).sema_, (c).waiters_count_, 0); \
-      MUTEX_UNLOCK((c).waiters_count_lock_); \
-      WaitForSingleObject((c).waiters_done_, INFINITE); \
-      (c).was_broadcast_ = 0; \
-    } \
-    else { \
-      MUTEX_UNLOCK((c).waiters_count_lock_); \
-    } \
-  } \
-} while (0)
-
-#  define COND_WAIT(c,m) do { \
-  MUTEX_LOCK((c).waiters_count_lock_); \
-  (c).waiters_count_++; \
-  MUTEX_UNLOCK((c).waiters_count_lock_); \
-  SignalObjectAndWait((m), (c).sema_, INFINITE, FALSE); \
-  MUTEX_LOCK((c).waiters_count_lock_); \
-  (c).waiters_count_--; \
-  { \
-    int last_waiter = (c).was_broadcast_ && (c).waiters_count_ == 0; \
-    MUTEX_UNLOCK((c).waiters_count_lock_); \
-    if (last_waiter)  { \
-      SignalObjectAndWait((c).waiters_done_, (m), INFINITE, FALSE); \
-    } \
-    else { \
-      WaitForSingleObject((m), FALSE); \
-    } \
-  } \
-} while (0)
-
-#  define COND_FIN(c) do { \
-  CloseHandle((c).waiters_done_); \
-  MUTEX_FIN((c).waiters_count_lock_); \
-  CloseHandle((c).sema_); \
-} while (0)
+typedef CONDITION_VARIABLE grn_cond;
+#  define COND_INIT(c) InitializeConditionVariable(&(c))
+#  define COND_SIGNAL(c) WakeConditionVariable(&(c))
+#  define COND_BROADCAST(c) WakeAllConditionVariable(&(c))
+#  define COND_WAIT(c,cs) SleepConditionVariableCS(&(c), &(cs), INFINITE)
+#  define COND_FIN(c)
 
 # else /* WIN32 */
 /* todo */
@@ -414,9 +332,6 @@ typedef int grn_cond;
 
 # endif /* WIN32 */
 
-# define MUTEX_INIT_SHARED MUTEX_INIT
-# define COND_INIT_SHARED COND_INIT
-
 # define GRN_TEST_YIELD() do {} while (0)
 
 #endif /* HAVE_PTHREAD_H */
@@ -442,21 +357,25 @@ typedef int grn_cond;
 # define GRN_FMT_INT32U PRIu32
 # define GRN_FMT_INT64D PRId64
 # define GRN_FMT_INT64U PRIu64
+# define GRN_FMT_INT64X PRIx64
 #else /* HAVE_INTTYPES_H */
 # ifdef WIN32
 #  define GRN_FMT_INT32D "I32d"
 #  define GRN_FMT_INT32U "I32u"
 #  define GRN_FMT_INT64D "I64d"
 #  define GRN_FMT_INT64U "I64u"
+#  define GRN_FMT_INT64X "I64x"
 # else /* WIN32 */
 #  define GRN_FMT_INT32D "d"
 #  define GRN_FMT_INT32U "u"
 #  ifdef __x86_64__
 #   define GRN_FMT_INT64D "ld"
 #   define GRN_FMT_INT64U "lu"
+#   define GRN_FMT_INT64X "lx"
 #  else /* __x86_64__ */
 #   define GRN_FMT_INT64D "lld"
 #   define GRN_FMT_INT64U "llu"
+#   define GRN_FMT_INT64X "llx"
 #  endif /* __x86_64__ */
 # endif /* WIN32 */
 #endif /* HAVE_INTTYPES_H */
