@@ -20,6 +20,7 @@
 
 #include "grn.h"
 #include "grn_ctx.h"
+#include "grn_options.h"
 #include "grn_store.h"
 #include "grn_rset.h"
 
@@ -49,6 +50,7 @@ struct _grn_db {
   grn_tiny_array values;
   grn_critical_section lock;
   grn_cache *cache;
+  grn_options *options;
 };
 
 #define GRN_SERIALIZED_SPEC_INDEX_SPEC   0
@@ -130,6 +132,24 @@ void grn_db_generate_pathname(grn_ctx *ctx,
                               char *buffer);
 grn_rc grn_db_clear_dirty(grn_ctx *ctx, grn_obj *db);
 
+grn_rc grn_db_set_option_values(grn_ctx *ctx,
+                                grn_obj *db,
+                                grn_id id,
+                                const char *name,
+                                int name_length,
+                                grn_obj *values);
+grn_option_revision
+grn_db_get_option_values(grn_ctx *ctx,
+                         grn_obj *db,
+                         grn_id id,
+                         const char *name,
+                         int name_length,
+                         grn_option_revision revision,
+                         grn_obj *values);
+grn_rc grn_db_clear_option_values(grn_ctx *ctx,
+                                  grn_obj *db,
+                                  grn_id id);
+
 grn_rc _grn_table_delete_by_id(grn_ctx *ctx, grn_obj *table, grn_id id,
                                grn_table_delete_optarg *optarg);
 
@@ -250,6 +270,11 @@ struct _grn_proc {
     struct {
       grn_command_run_func *run;
     } command;
+    struct {
+      grn_tokenizer_init_func *init;
+      grn_tokenizer_next_func *next;
+      grn_tokenizer_fin_func  *fin;
+    } tokenizer;
     struct {
       grn_token_filter_init_func   *init;
       grn_token_filter_filter_func *filter;
@@ -365,42 +390,6 @@ grn_rc grn_db_init_builtin_types(grn_ctx *ctx);
 
 typedef struct _grn_expr grn_expr;
 
-#define GRN_EXPR_CODE_RELATIONAL_EXPRESSION (0x01)
-
-typedef struct {
-  grn_obj *value;
-  int32_t nargs;
-  grn_operator op;
-  uint8_t flags;
-  int32_t modify;
-} grn_expr_code;
-
-#define GRN_EXPR_CONST_BLK_SIZE GRN_STACK_SIZE
-
-struct _grn_expr {
-  grn_db_obj obj;
-  grn_obj name_buf;
-  grn_expr_var *vars;
-  uint32_t nvars;
-  /* -- compatible with grn_proc -- */
-
-  uint16_t cacheable;
-  uint16_t taintable;
-  grn_obj **const_blks;
-  grn_obj *values;
-  grn_expr_code *codes;
-  uint32_t nconsts;
-  uint32_t values_curr;
-  uint32_t values_tail;
-  uint32_t values_size;
-  uint32_t codes_curr;
-  uint32_t codes_size;
-
-  grn_obj objs;
-  grn_obj dfi;
-  grn_expr_code *code0;
-};
-
 grn_rc grn_expr_parser_close(grn_ctx *ctx);
 
 /**
@@ -463,6 +452,9 @@ void grn_obj_spec_save(grn_ctx *ctx, grn_db_obj *obj);
 
 grn_rc grn_obj_reinit_for(grn_ctx *ctx, grn_obj *obj, grn_obj *domain_obj);
 
+void grn_obj_ensure_bulk(grn_ctx *ctx, grn_obj *obj);
+void grn_obj_ensure_vector(grn_ctx *ctx, grn_obj *obj);
+
 void grn_expr_pack(grn_ctx *ctx, grn_obj *buf, grn_obj *expr);
 GRN_API grn_rc grn_expr_inspect(grn_ctx *ctx, grn_obj *buf, grn_obj *expr);
 grn_hash *grn_expr_get_vars(grn_ctx *ctx, grn_obj *expr, unsigned int *nvars);
