@@ -19,10 +19,10 @@
 #pragma once
 
 #include "grn_ctx.h"
+#include "grn_token.h"
+#include "grn_tokenizer.h"
 #include "grn_db.h"
 
-#include <groonga/tokenizer.h>
-
 #ifdef __cplusplus
 extern "C" {
 #endif
@@ -39,11 +39,6 @@ typedef enum {
   GRN_TOKEN_CURSOR_NOT_FOUND
 } grn_token_cursor_status;
 
-struct _grn_token {
-  grn_obj data;
-  grn_token_status status;
-};
-
 typedef struct {
   grn_obj *table;
   const unsigned char *orig;
@@ -56,14 +51,19 @@ typedef struct {
   grn_bool force_prefix;
   grn_obj_flags table_flags;
   grn_encoding encoding;
-  grn_obj *tokenizer;
-  grn_proc_ctx pctx;
+  struct {
+    grn_obj *object;
+    grn_proc_ctx pctx;
+    grn_tokenizer_query query;
+    void *user_data;
+    grn_token current_token;
+    grn_token next_token;
+  } tokenizer;
   struct {
     grn_obj *objects;
     void **data;
   } token_filter;
   uint32_t variant;
-  grn_obj *nstr;
 } grn_token_cursor;
 
 #define GRN_TOKEN_CURSOR_ENABLE_TOKENIZED_DELIMITER (0x01L<<0)
@@ -76,6 +76,9 @@ GRN_API grn_token_cursor *grn_token_cursor_open(grn_ctx *ctx, grn_obj *table,
 GRN_API grn_id grn_token_cursor_next(grn_ctx *ctx, grn_token_cursor *token_cursor);
 GRN_API grn_rc grn_token_cursor_close(grn_ctx *ctx, grn_token_cursor *token_cursor);
 
+GRN_API grn_token *
+grn_token_cursor_get_token(grn_ctx *ctx, grn_token_cursor *token_cursor);
+
 #ifdef __cplusplus
 }
 #endif
