@@ -1,6 +1,7 @@
 /* -*- c-basic-offset: 2 -*- */
 /*
   Copyright(C) 2018 Brazil
+  Copyright(C) 2018 Kouhei Sutou <kou@clear-code.com>
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -65,6 +66,8 @@ struct _grn_highlighter {
   struct {
     grn_obj *object;
     grn_encoding encoding;
+    grn_obj lazy_keywords;
+    grn_obj lazy_keyword_ids;
     grn_obj *token_id_chunks;
     grn_obj token_id_chunk_ids;
     grn_obj token_id_chunk;
@@ -111,6 +114,10 @@ grn_highlighter_open(grn_ctx *ctx)
 
   highlighter->lexicon.object = NULL;
   highlighter->lexicon.encoding = GRN_ENC_NONE;
+  GRN_TEXT_INIT(&(highlighter->lexicon.lazy_keywords), GRN_OBJ_VECTOR);
+  GRN_RECORD_INIT(&(highlighter->lexicon.lazy_keyword_ids),
+                  GRN_OBJ_VECTOR,
+                  GRN_ID_NIL);
   highlighter->lexicon.token_id_chunks = NULL;
   GRN_RECORD_INIT(&(highlighter->lexicon.token_id_chunk_ids),
                   GRN_OBJ_VECTOR,
@@ -141,6 +148,8 @@ grn_highlighter_close(grn_ctx *ctx,
     grn_obj_close(ctx, highlighter->pat.keywords);
   }
 
+  GRN_OBJ_FIN(ctx, &(highlighter->lexicon.lazy_keywords));
+  GRN_OBJ_FIN(ctx, &(highlighter->lexicon.lazy_keyword_ids));
   if (highlighter->lexicon.token_id_chunks) {
     grn_obj_close(ctx, highlighter->lexicon.token_id_chunks);
   }
@@ -210,12 +219,17 @@ grn_highlighter_prepare_lexicon(grn_ctx *ctx,
                                 grn_highlighter *highlighter)
 {
   grn_bool have_token_id_chunks = GRN_FALSE;
+  grn_obj *lazy_keywords = &(highlighter->lexicon.lazy_keywords);
+  grn_id lexicon_id;
   grn_obj *token_id_chunk_ids = &(highlighter->lexicon.token_id_chunk_ids);
   size_t i, n_keywords;
   grn_obj *token_id_chunk = &(highlighter->lexicon.token_id_chunk);
 
-  highlighter->lexicon.token_ids.header.domain =
-    grn_obj_id(ctx, highlighter->lexicon.object);
+  GRN_BULK_REWIND(lazy_keywords);
+
+  lexicon_id = grn_obj_id(ctx, highlighter->lexicon.object);
+  highlighter->lexicon.lazy_keyword_ids.header.domain = lexicon_id;
+  highlighter->lexicon.token_ids.header.domain = lexicon_id;
 
   grn_table_get_info(ctx,
                      highlighter->lexicon.object,
@@ -293,6 +307,21 @@ grn_highlighter_prepare_lexicon(grn_ctx *ctx,
     GRN_BULK_REWIND(token_id_chunk);
     while ((token_id = grn_token_cursor_next(ctx, cursor)) != GRN_ID_NIL) {
       GRN_TEXT_PUT(ctx, token_id_chunk, &token_id, sizeof(grn_id));
+      if (cursor->force_prefix &&
+          highlighter->lexicon.object->header.type != GRN_TABLE_HASH_KEY) {
+        grn_token *token;
+        const char *data;
+        size_t data_length;
+
+        token = grn_token_cursor_get_token(ctx, cursor);
+        data = grn_token_get_data_raw(ctx, token, &data_length);
+        grn_vector_add_element(ctx,
+                               lazy_keywords,
+                               data,
+                               data_length,
+                               0,
+                               GRN_DB_TEXT);
+      }
     }
     grn_token_cursor_close(ctx, cursor);
 
@@ -418,8 +447,49 @@ grn_highlighter_prepare(grn_ctx *ctx,
   highlighter->need_prepared = GRN_FALSE;
 }
 
+static grn_bool
+grn_ids_is_included(grn_id *ids, size_t n_ids, grn_id id)
+{
+  size_t i;
+
+  for (i = 0; i < n_ids; i++) {
+    if (ids[i] == ids[0]) {
+      return GRN_TRUE;
+    }
+  }
+
+  return GRN_FALSE;
+}
+
+static void
+grn_highlighter_log_location(grn_ctx *ctx,
+                             grn_log_level level,
+                             const char *tag,
+                             grn_highlighter_location *location,
+                             const char *text,
+                             size_t text_length)
+{
+  if (!grn_logger_pass(ctx, level)) {
+    return;
+  }
+
+  GRN_LOG(ctx,
+          level,
+          "%s[location] "
+          "[%" GRN_FMT_INT64U "...%" GRN_FMT_INT64U "](%u)[%s] <%.*s>",
+          tag,
+          location->offset,
+          location->offset + location->length,
+          location->first_character_length,
+          location->have_overlap ? "overlapped" : "not-overlapped",
+          (int)location->length,
+          text + location->offset);
+}
+
 static uint64_t
 grn_highlighter_highlight_lexicon_flush(grn_ctx *ctx,
+                                        grn_log_level log_level,
+                                        const char *tag,
                                         grn_highlighter *highlighter,
                                         const char *text,
                                         size_t text_length,
@@ -427,6 +497,7 @@ grn_highlighter_highlight_lexicon_flush(grn_ctx *ctx,
                                         grn_highlighter_location *location,
                                         uint64_t offset)
 {
+  grn_highlighter_log_location(ctx, log_level, tag, location, text, text_length);
   if (location->offset > offset) {
     grn_text_escape_xml(ctx,
                         output,
@@ -455,11 +526,15 @@ grn_highlighter_highlight_lexicon(grn_ctx *ctx,
                                   size_t text_length,
                                   grn_obj *output)
 {
+  const char *tag = "[highlighter][highlight][lexicon]";
+  grn_log_level log_level = GRN_LOG_DEBUG;
   grn_token_cursor *cursor;
+  grn_obj *lazy_keyword_ids = &(highlighter->lexicon.lazy_keyword_ids);
   grn_obj *token_ids = &(highlighter->lexicon.token_ids);
   grn_obj *token_locations = &(highlighter->lexicon.token_locations);
   grn_obj *candidates = &(highlighter->lexicon.candidates);
 
+  GRN_BULK_REWIND(lazy_keyword_ids);
   GRN_BULK_REWIND(token_ids);
   GRN_BULK_REWIND(token_locations);
   cursor = grn_token_cursor_open(ctx,
@@ -470,11 +545,13 @@ grn_highlighter_highlight_lexicon(grn_ctx *ctx,
                                  0);
   if (!cursor) {
     ERR(ctx->rc,
-        "[highlighter][highlight][lexicon] failed to start tokenizing: %s",
+        "%s failed to start tokenizing: %s",
+        tag,
         ctx->errbuf);
     return;
   }
 
+  GRN_LOG(ctx, log_level, "%s[tokenize][start]", tag);
   while (cursor->status == GRN_TOKEN_CURSOR_DOING) {
     grn_id token_id = grn_token_cursor_next(ctx, cursor);
     grn_highlighter_location location;
@@ -491,11 +568,89 @@ grn_highlighter_highlight_lexicon(grn_ctx *ctx,
       grn_token_get_source_first_character_length(ctx, token);
     location.have_overlap = grn_token_have_overlap(ctx, token);
     GRN_TEXT_PUT(ctx, token_locations, &location, sizeof(location));
+    grn_highlighter_log_location(ctx,
+                                 log_level,
+                                 tag,
+                                 &location,
+                                 text,
+                                 text_length);
   }
   grn_token_cursor_close(ctx, cursor);
+  GRN_LOG(ctx, log_level, "%s[tokenize][end]", tag);
+
+  {
+    grn_obj *lexicon = highlighter->lexicon.object;
+    grn_obj *chunks = highlighter->lexicon.token_id_chunks;
+    grn_obj *lazy_keywords = &(highlighter->lexicon.lazy_keywords);
+    size_t i;
+    size_t n_keywords;
+
+    n_keywords = grn_vector_size(ctx, lazy_keywords);
+    for (i = 0; i < n_keywords; i++) {
+      const char *keyword;
+      unsigned int keyword_length;
+
+      keyword_length = grn_vector_get_element(ctx,
+                                              lazy_keywords,
+                                              i,
+                                              &keyword,
+                                              NULL,
+                                              NULL);
+      GRN_LOG(ctx,
+              log_level,
+              "%s[prefix-search][start] %" GRN_FMT_SIZE ":<%.*s>",
+              tag,
+              i,
+              (int)keyword_length,
+              keyword);
+      GRN_TABLE_EACH_BEGIN_MIN(ctx,
+                               lexicon,
+                               cursor,
+                               id,
+                               keyword, keyword_length,
+                               GRN_CURSOR_PREFIX) {
+        int added = 0;
+
+        {
+          grn_encoding encoding = ctx->encoding;
+          ctx->encoding = GRN_ENC_NONE;
+          grn_table_add(ctx, chunks, &id, sizeof(grn_id), &added);
+          ctx->encoding = encoding;
+        }
+        if (grn_logger_pass(ctx, GRN_LOG_DEBUG)) {
+          void *key;
+          int key_size;
+          key_size = grn_table_cursor_get_key(ctx, cursor, &key);
+          GRN_LOG(ctx,
+                  log_level,
+                  "%s[prefix-search][%s] %" GRN_FMT_SIZE ":<%.*s>:<%.*s>",
+                  tag,
+                  added ? "added" : "not-added",
+                  i,
+                  (int)keyword_length,
+                  keyword,
+                  key_size,
+                  (const char *)key);
+        }
+        if (added) {
+          GRN_RECORD_PUT(ctx, lazy_keyword_ids, id);
+        }
+      } GRN_TABLE_EACH_END(ctx, cursor);
+      GRN_LOG(ctx,
+              log_level,
+              "%s[prefix-search][end] %" GRN_FMT_SIZE ":<%.*s>",
+              tag,
+              i,
+              (int)keyword_length,
+              keyword);
+    }
+  }
 
   GRN_BULK_REWIND(candidates);
   {
+    grn_obj *lazy_keyword_ids = &(highlighter->lexicon.lazy_keyword_ids);
+    grn_id *raw_lazy_keyword_ids;
+    size_t n_lazy_keyword_ids;
     grn_pat *chunks = (grn_pat *)(highlighter->lexicon.token_id_chunks);
     size_t i;
     size_t n_token_ids = GRN_BULK_VSIZE(token_ids) / sizeof(grn_id);
@@ -503,9 +658,17 @@ grn_highlighter_highlight_lexicon(grn_ctx *ctx,
     grn_highlighter_location *raw_token_locations =
       (grn_highlighter_location *)GRN_BULK_HEAD(token_locations);
 
+    raw_lazy_keyword_ids = (grn_id *)GRN_BULK_HEAD(lazy_keyword_ids);
+    n_lazy_keyword_ids = GRN_BULK_VSIZE(lazy_keyword_ids) / sizeof(grn_id);
+
     for (i = 0; i < n_token_ids; i++) {
       grn_id chunk_id;
 
+      GRN_LOG(ctx,
+              log_level,
+              "%s[lcp-search][start] %" GRN_FMT_SIZE,
+              tag,
+              i);
       {
         grn_encoding encoding = ctx->encoding;
         /* token_id_chunk is a binary data */
@@ -517,20 +680,39 @@ grn_highlighter_highlight_lexicon(grn_ctx *ctx,
         ctx->encoding = encoding;
       }
       if (chunk_id == GRN_ID_NIL) {
+        GRN_LOG(ctx,
+                log_level,
+                "%s[lcp-search][end][nonexistent] %" GRN_FMT_SIZE,
+                tag,
+                i);
         continue;
       }
 
       {
+        grn_id *ids;
         uint32_t key_size;
         size_t j;
         size_t n_ids;
         grn_highlighter_location candidate;
         grn_highlighter_location *first = raw_token_locations + i;
 
-        _grn_pat_key(ctx, chunks, chunk_id, &key_size);
+        candidate.have_overlap = GRN_FALSE;
+        candidate.first_character_length = 0;
+
+        {
+          grn_encoding encoding = ctx->encoding;
+          ctx->encoding = GRN_ENC_NONE;
+          ids = (grn_id *)_grn_pat_key(ctx, chunks, chunk_id, &key_size);
+          ctx->encoding = encoding;
+        }
         n_ids = key_size / sizeof(grn_id);
         candidate.offset = first->offset;
-        if (first->have_overlap && n_ids > 1) {
+        if (n_ids == 1 &&
+            grn_ids_is_included(raw_lazy_keyword_ids,
+                                n_lazy_keyword_ids,
+                                ids[0])) {
+          candidate.length = first->first_character_length;
+        } else if (first->have_overlap && n_ids > 1) {
           candidate.length = first->first_character_length;
         } else {
           candidate.length = first->length;
@@ -559,11 +741,36 @@ grn_highlighter_highlight_lexicon(grn_ctx *ctx,
           }
         }
         GRN_TEXT_PUT(ctx, candidates, &candidate, sizeof(candidate));
+        grn_highlighter_log_location(ctx,
+                                     log_level,
+                                     tag,
+                                     &candidate,
+                                     text,
+                                     text_length);
+        GRN_LOG(ctx,
+                log_level,
+                "%s[lcp-search][end] %" GRN_FMT_SIZE,
+                tag,
+                i);
         i += n_ids - 1;
       }
     }
   }
 
+  {
+    grn_obj *chunks = highlighter->lexicon.token_id_chunks;
+    size_t i, n_ids;
+
+    n_ids = GRN_BULK_VSIZE(lazy_keyword_ids) / sizeof(grn_id);
+    for (i = 0; i < n_ids; i++) {
+      grn_id id = GRN_RECORD_VALUE_AT(lazy_keyword_ids, i);
+      grn_encoding encoding = ctx->encoding;
+      ctx->encoding = GRN_ENC_NONE;
+      grn_table_delete(ctx, chunks, &id, sizeof(grn_id));
+      ctx->encoding = encoding;
+    }
+  }
+
   {
     size_t i;
     size_t n_candidates =
@@ -571,6 +778,11 @@ grn_highlighter_highlight_lexicon(grn_ctx *ctx,
     grn_highlighter_location *raw_candidates =
       (grn_highlighter_location *)GRN_BULK_HEAD(candidates);
 
+    GRN_LOG(ctx,
+            log_level,
+            "%s[highlight][start] %" GRN_FMT_SIZE,
+            tag,
+            n_candidates);
     if (n_candidates == 0) {
       grn_text_escape_xml(ctx, output, text, text_length);
     } else {
@@ -597,6 +809,8 @@ grn_highlighter_highlight_lexicon(grn_ctx *ctx,
         }
         if (current->offset > previous->offset) {
           offset = grn_highlighter_highlight_lexicon_flush(ctx,
+                                                           log_level,
+                                                           tag,
                                                            highlighter,
                                                            text,
                                                            text_length,
@@ -607,6 +821,8 @@ grn_highlighter_highlight_lexicon(grn_ctx *ctx,
         previous = current;
       }
       offset = grn_highlighter_highlight_lexicon_flush(ctx,
+                                                       log_level,
+                                                       tag,
                                                        highlighter,
                                                        text,
                                                        text_length,
@@ -620,6 +836,11 @@ grn_highlighter_highlight_lexicon(grn_ctx *ctx,
                             text_length - offset);
       }
     }
+    GRN_LOG(ctx,
+            log_level,
+            "%s[highlight][end] %" GRN_FMT_SIZE,
+            tag,
+            n_candidates);
   }
 }
 
@@ -802,3 +1023,17 @@ grn_highlighter_add_keyword(grn_ctx *ctx,
 exit :
   GRN_API_RETURN(ctx->rc);
 }
+
+grn_rc
+grn_highlighter_clear_keywords(grn_ctx *ctx,
+                               grn_highlighter *highlighter)
+{
+  grn_obj *raw_keywords = &(highlighter->raw_keywords);
+
+  GRN_API_ENTER;
+
+  GRN_BULK_REWIND(raw_keywords);
+  highlighter->need_prepared = GRN_TRUE;
+
+  GRN_API_RETURN(ctx->rc);
+}
