@@ -1,6 +1,7 @@
 /* -*- c-basic-offset: 2 -*- */
 /*
   Copyright(C) 2009-2018 Brazil
+  Copyright(C) 2018 Kouhei Sutou <kou@clear-code.com>
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -51,10 +52,10 @@
 
 #ifndef WIN32
 # ifdef HAVE_FUTIMENS
-#  define grn_futimens(fd, times) futimens((fd), (times))
+#  define grn_futimens(fd, time_values, time_specs) futimens((fd), (time_specs))
 #  define grn_futimens_name "futimens"
 # elif defined(HAVE_FUTIMES) /* HAVE_FUTIMENS */
-#  define grn_futimens(fd, times) futimes((fd), (times))
+#  define grn_futimens(fd, time_values, time_specs) futimes((fd), (time_values))
 #  define grn_futimens_name "futimes"
 # else /* HAVE_FUTIMENS */
 #  error Support for either the futimens() or futimens() function is required
@@ -1741,8 +1742,17 @@ grn_fileinfo_open_common(grn_ctx *ctx, fileinfo *fi, const char *path, int flags
 {
   /* may be wrong if flags is just only O_RDWR */
   if ((flags & O_CREAT)) {
+    grn_bool exist = GRN_FALSE;
     DWORD dwCreationDisposition;
     const char *flags_description;
+
+    {
+      struct stat stat_buffer;
+      if (stat(path, &stat_buffer) == 0) {
+        exist = GRN_TRUE;
+      }
+    }
+
     if (flags & O_EXCL) {
       dwCreationDisposition = CREATE_NEW;
       flags_description = "O_RDWR|O_CREAT|O_EXCL";
@@ -1757,6 +1767,16 @@ grn_fileinfo_open_common(grn_ctx *ctx, fileinfo *fi, const char *path, int flags
     if (fi->fh == INVALID_HANDLE_VALUE) {
       SERR("CreateFile(<%s>, <%s>) failed",
            path, flags_description);
+      if (!exist) {
+        struct stat stat_buffer;
+        if (stat(path, &stat_buffer) == 0) {
+          GRN_LOG(ctx, GRN_LOG_INFO,
+                  "[io][open] "
+                  "delete a newly created file because of open error: <%s>",
+                  path);
+          DeleteFile(path);
+        }
+      }
       goto exit;
     }
 
@@ -2150,10 +2170,14 @@ grn_msync(grn_ctx *ctx, fileinfo *fi, void *start, size_t length)
   }
 
   if (fi->fd > 0) {
-    struct timeval times[2];
-    gettimeofday(&(times[0]), NULL);
-    times[1] = times[0];
-    r = grn_futimens(fi->fd, times);
+    struct timeval time_values[2];
+    struct timespec time_specs[2];
+    gettimeofday(&(time_values[0]), NULL);
+    time_values[1] = time_values[0];
+    time_specs[0].tv_sec = time_values[0].tv_sec;
+    time_specs[0].tv_nsec = time_values[0].tv_usec * 1000;
+    time_specs[1] = time_specs[0];
+    r = grn_futimens(fi->fd, time_values, time_specs);
     if (r == -1) {
       SERR("%s failed: <%d>", grn_futimens_name, fi->fd);
     }
