@@ -1,6 +1,7 @@
 /* -*- c-basic-offset: 2 -*- */
 /*
   Copyright(C) 2009-2017 Brazil
+  Copyright(C) 2018 Kouhei Sutou <kou@clear-code.com>
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -362,6 +363,12 @@ grn_logger_set(grn_ctx *ctx, const grn_logger *logger)
   return GRN_SUCCESS;
 }
 
+grn_bool
+grn_logger_is_default_logger(grn_ctx *ctx)
+{
+  return current_logger.log == default_logger.log;
+}
+
 void
 grn_logger_set_max_level(grn_ctx *ctx, grn_log_level max_level)
 {
@@ -426,11 +433,36 @@ grn_logger_putv(grn_ctx *ctx,
     if (current_logger.flags & GRN_LOG_LOCATION) {
       grn_snprintf(lbuf, LBUFSIZE, LBUFSIZE,
                    "%d %s:%d %s()", grn_getpid(), file, line, func);
-    } else if (current_logger.flags & GRN_LOG_PID) {
-      grn_snprintf(lbuf, LBUFSIZE, LBUFSIZE,
-                   "%d", grn_getpid());
     } else {
+      char *lbuf_current = lbuf;
+      size_t lbuf_rest_size = LBUFSIZE;
       lbuf[0] = '\0';
+      lbuf_current = lbuf;
+      if (current_logger.flags & GRN_LOG_PID) {
+        size_t lbuf_size;
+        grn_snprintf(lbuf_current, lbuf_rest_size, lbuf_rest_size,
+                     "%d", grn_getpid());
+        lbuf_size = strlen(lbuf_current);
+        lbuf_current += lbuf_size;
+        lbuf_rest_size -= lbuf_size;
+      }
+      if (current_logger.flags & GRN_LOG_THREAD_ID) {
+        const char *prefix = "";
+        size_t lbuf_size;
+        if (lbuf_current != lbuf) {
+          prefix = "|";
+        }
+#ifdef HAVE_PTHREAD_H
+        grn_snprintf(lbuf_current, lbuf_rest_size, lbuf_rest_size,
+                     "%s%08x", prefix, (uint32_t)pthread_self());
+#elif defined(WIN32) /* HAVE_PTHREAD_H */
+        grn_snprintf(lbuf_current, lbuf_rest_size, lbuf_rest_size,
+                     "%s%p", prefix, GetCurrentThread());
+#endif /* HAVE_PTHREAD_H */
+        lbuf_size = strlen(lbuf_current);
+        lbuf_current += lbuf_size;
+        lbuf_rest_size -= lbuf_size;
+      }
     }
     current_logger.log(ctx, level, tbuf, "", mbuf, lbuf,
                        current_logger.user_data);
