@@ -1,6 +1,6 @@
 /* -*- c-basic-offset: 2 -*- */
 /*
-  Copyright(C) 2009-2012 Brazil
+  Copyright(C) 2009-2018 Brazil
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -16,15 +16,66 @@
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
-#include "groonga_in.h"
+#include "grn.h"
 #include <string.h>
-#include "string_in.h"
-#include "normalizer_in.h"
-#include "str.h"
-#include "util.h"
+#include "grn_string.h"
+#include "grn_normalizer.h"
+#include "grn_str.h"
+#include "grn_util.h"
 
 #include <groonga/tokenizer.h>
 
+const char *
+grn_char_type_to_string(grn_char_type type)
+{
+  const char *string = "unknown";
+
+#define CHAR_TYPE_STRING_WITH_BLANK(type_string) do {   \
+    if (GRN_CHAR_IS_BLANK(type)) {                      \
+      string = type_string "|blank";                    \
+    } else {                                            \
+      string = type_string;                             \
+    }                                                   \
+  } while (GRN_FALSE)
+
+  switch (GRN_CHAR_TYPE(type)) {
+  case GRN_CHAR_NULL :
+    CHAR_TYPE_STRING_WITH_BLANK("null");
+    break;
+  case GRN_CHAR_ALPHA :
+    CHAR_TYPE_STRING_WITH_BLANK("alpha");
+    break;
+  case GRN_CHAR_DIGIT :
+    CHAR_TYPE_STRING_WITH_BLANK("digit");
+    break;
+  case GRN_CHAR_SYMBOL :
+    CHAR_TYPE_STRING_WITH_BLANK("symbol");
+    break;
+  case GRN_CHAR_HIRAGANA :
+    CHAR_TYPE_STRING_WITH_BLANK("hiragana");
+    break;
+  case GRN_CHAR_KATAKANA :
+    CHAR_TYPE_STRING_WITH_BLANK("katakana");
+    break;
+  case GRN_CHAR_KANJI :
+    CHAR_TYPE_STRING_WITH_BLANK("kanji");
+    break;
+  case GRN_CHAR_OTHERS :
+    CHAR_TYPE_STRING_WITH_BLANK("others");
+    break;
+  case GRN_CHAR_EMOJI :
+    CHAR_TYPE_STRING_WITH_BLANK("emoji");
+    break;
+  default :
+    CHAR_TYPE_STRING_WITH_BLANK("unknown");
+    break;
+  }
+
+#undef CHAR_TYPE_STRING_WITH_BLANK
+
+  return string;
+}
+
 static grn_string *
 grn_fake_string_open(grn_ctx *ctx, grn_string *string)
 {
@@ -54,7 +105,7 @@ grn_fake_string_open(grn_ctx *ctx, grn_string *string)
       if (!grn_tokenizer_is_tokenized_delimiter(ctx,
                                                 source_current, char_length,
                                                 ctx->encoding)) {
-        memcpy(destination, source_current, char_length);
+        grn_memcpy(destination, source_current, char_length);
         destination += char_length;
         destination_length += char_length;
       }
@@ -63,7 +114,7 @@ grn_fake_string_open(grn_ctx *ctx, grn_string *string)
     nstr->normalized[destination_length] = '\0';
     nstr->normalized_length_in_bytes = destination_length;
   } else {
-    memcpy(nstr->normalized, str, str_len);
+    grn_memcpy(nstr->normalized, str, str_len);
     nstr->normalized[str_len] = '\0';
     nstr->normalized_length_in_bytes = str_len;
   }
@@ -129,8 +180,12 @@ grn_fake_string_open(grn_ctx *ctx, grn_string *string)
 }
 
 grn_obj *
-grn_string_open_(grn_ctx *ctx, const char *str, unsigned int str_len,
-                 grn_obj *normalizer, int flags, grn_encoding encoding)
+grn_string_open_(grn_ctx *ctx,
+                 const char *str,
+                 unsigned int str_len,
+                 grn_obj *lexicon_or_normalizer,
+                 int flags,
+                 grn_encoding encoding)
 {
   grn_string *string;
   grn_obj *obj;
@@ -140,8 +195,21 @@ grn_string_open_(grn_ctx *ctx, const char *str, unsigned int str_len,
     return NULL;
   }
 
+  is_normalizer_auto = (lexicon_or_normalizer == GRN_NORMALIZER_AUTO);
+  if (is_normalizer_auto) {
+    lexicon_or_normalizer = grn_ctx_get(ctx, GRN_NORMALIZER_AUTO_NAME, -1);
+    if (!lexicon_or_normalizer) {
+      ERR(GRN_INVALID_ARGUMENT,
+          "[string][open] NormalizerAuto normalizer isn't available");
+      return NULL;
+    }
+  }
+
   string = GRN_MALLOCN(grn_string, 1);
   if (!string) {
+    if (is_normalizer_auto) {
+      grn_obj_unlink(ctx, lexicon_or_normalizer);
+    }
     GRN_LOG(ctx, GRN_LOG_ALERT,
             "[string][open] failed to allocate memory");
     return NULL;
@@ -156,37 +224,58 @@ grn_string_open_(grn_ctx *ctx, const char *str, unsigned int str_len,
   string->n_characters = 0;
   string->checks = NULL;
   string->ctypes = NULL;
+  string->offsets = NULL;
   string->encoding = encoding;
   string->flags = flags;
+  string->lexicon = NULL;
 
-  if (!normalizer) {
+  if (!lexicon_or_normalizer) {
     return (grn_obj *)grn_fake_string_open(ctx, string);
   }
 
-  is_normalizer_auto = (normalizer == GRN_NORMALIZER_AUTO);
-  if (is_normalizer_auto) {
-    normalizer = grn_ctx_at(ctx, GRN_DB_NORMALIZER_AUTO);
+  {
+    grn_obj *normalizer = NULL;
+    if (lexicon_or_normalizer) {
+      if (grn_obj_is_table(ctx, lexicon_or_normalizer)) {
+        string->lexicon = lexicon_or_normalizer;
+        normalizer = grn_obj_get_info(ctx,
+                                      string->lexicon,
+                                      GRN_INFO_NORMALIZER,
+                                      NULL);
+      } else {
+        normalizer = lexicon_or_normalizer;
+      }
+    }
+    if (!normalizer) {
+      return (grn_obj *)grn_fake_string_open(ctx, string);
+    }
+    grn_normalizer_normalize(ctx, normalizer, (grn_obj *)string);
   }
-
-  /* TODO: check rc */
-  grn_normalizer_normalize(ctx, normalizer, (grn_obj *)string);
   if (ctx->rc) {
     grn_obj_close(ctx, obj);
     obj = NULL;
   }
 
   if (is_normalizer_auto) {
-    grn_obj_unlink(ctx, normalizer);
+    grn_obj_unlink(ctx, lexicon_or_normalizer);
   }
 
   return obj;
 }
 
 grn_obj *
-grn_string_open(grn_ctx *ctx, const char *str, unsigned int str_len,
-                grn_obj *normalizer, int flags)
+grn_string_open(grn_ctx *ctx,
+                const char *str,
+                unsigned int str_len,
+                grn_obj *lexicon_or_normalizer,
+                int flags)
 {
-  return grn_string_open_(ctx, str, str_len, normalizer, flags, ctx->encoding);
+  return grn_string_open_(ctx,
+                          str,
+                          str_len,
+                          lexicon_or_normalizer,
+                          flags,
+                          ctx->encoding);
 }
 
 grn_rc
@@ -238,6 +327,9 @@ grn_string_get_normalized(grn_ctx *ctx, grn_obj *string,
     if (n_characters) { *n_characters = string_->n_characters; }
     rc = GRN_SUCCESS;
   } else {
+    if (normalized) { *normalized = NULL; }
+    if (length_in_bytes) { *length_in_bytes = 0; }
+    if (n_characters) { *n_characters = 0; }
     rc = GRN_INVALID_ARGUMENT;
   }
   GRN_API_RETURN(rc);
@@ -323,6 +415,36 @@ grn_string_set_types(grn_ctx *ctx, grn_obj *string, unsigned char *types)
   GRN_API_RETURN(rc);
 }
 
+const uint64_t *
+grn_string_get_offsets(grn_ctx *ctx, grn_obj *string)
+{
+  uint64_t *offsets = NULL;
+  grn_string *string_ = (grn_string *)string;
+  GRN_API_ENTER;
+  if (string_) {
+    offsets = string_->offsets;
+  } else {
+    offsets = NULL;
+  }
+  GRN_API_RETURN(offsets);
+}
+
+grn_rc
+grn_string_set_offsets(grn_ctx *ctx, grn_obj *string, uint64_t *offsets)
+{
+  grn_rc rc;
+  grn_string *string_ = (grn_string *)string;
+  GRN_API_ENTER;
+  if (string_) {
+    if (string_->offsets) { GRN_FREE(string_->offsets); }
+    string_->offsets = offsets;
+    rc = GRN_SUCCESS;
+  } else {
+    rc = GRN_INVALID_ARGUMENT;
+  }
+  GRN_API_RETURN(rc);
+}
+
 grn_encoding
 grn_string_get_encoding(grn_ctx *ctx, grn_obj *string)
 {
@@ -335,6 +457,18 @@ grn_string_get_encoding(grn_ctx *ctx, grn_obj *string)
   GRN_API_RETURN(encoding);
 }
 
+grn_obj *
+grn_string_get_table(grn_ctx *ctx, grn_obj *string)
+{
+  grn_obj *table = NULL;
+  grn_string *string_ = (grn_string *)string;
+  GRN_API_ENTER;
+  if (string_) {
+    table = string_->lexicon;
+  }
+  GRN_API_RETURN(table);
+}
+
 grn_rc
 grn_string_inspect(grn_ctx *ctx, grn_obj *buffer, grn_obj *string)
 {
@@ -397,6 +531,7 @@ grn_string_close(grn_ctx *ctx, grn_obj *string)
     if (string_->normalized) { GRN_FREE(string_->normalized); }
     if (string_->ctypes) { GRN_FREE(string_->ctypes); }
     if (string_->checks) { GRN_FREE(string_->checks); }
+    if (string_->offsets) { GRN_FREE(string_->offsets); }
     GRN_FREE(string);
     rc = GRN_SUCCESS;
   } else {
