@@ -1,6 +1,6 @@
 /* -*- c-basic-offset: 2 -*- */
 /*
-  Copyright(C) 2012-2016 Brazil
+  Copyright(C) 2012-2017 Brazil
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -54,6 +54,7 @@ typedef struct {
   char *name;
   grn_obj *database;
   grn_cache *cache;
+  ngx_str_t cache_base_path;
 } ngx_http_groonga_loc_conf_t;
 
 typedef struct {
@@ -84,8 +85,7 @@ typedef void (*ngx_http_groonga_loc_conf_callback_pt)(ngx_http_groonga_loc_conf_
 
 ngx_module_t ngx_http_groonga_module;
 
-static grn_ctx ngx_http_groonga_context;
-static grn_ctx *context = &ngx_http_groonga_context;
+static grn_ctx *context = NULL;
 static ngx_http_groonga_loc_conf_t *ngx_http_groonga_current_location_conf = NULL;
 
 static char *
@@ -332,7 +332,9 @@ ngx_http_groonga_context_init(ngx_http_groonga_loc_conf_t *location_conf,
   }
 
   grn_ctx_use(context, location_conf->database);
-  grn_cache_current_set(context, location_conf->cache);
+  grn_db_set_cache(context,
+                   location_conf->database,
+                   location_conf->cache);
 
   /* TODO: It doesn't work yet. We need to implement request timeout
    * handler. */
@@ -527,7 +529,8 @@ ngx_http_groonga_context_receive_handler_typed(grn_ctx *context,
       context->stat |= GRN_CTX_QUIT;
     } else {
       context->rc = GRN_OPERATION_NOT_PERMITTED;
-      GRN_TEXT_PUTS(context, &(data->typed.body), "false");
+      result = "false";
+      result_size = strlen(result);
       context->stat &= ~GRN_CTX_QUIT;
     }
   }
@@ -1198,6 +1201,8 @@ ngx_http_groonga_create_loc_conf(ngx_conf_t *cf)
   conf->config_file = NULL;
   conf->config_line = 0;
   conf->cache = NULL;
+  conf->cache_base_path.data = NULL;
+  conf->cache_base_path.len = 0;
 
   return conf;
 }
@@ -1253,6 +1258,10 @@ ngx_http_groonga_merge_loc_conf(ngx_conf_t *cf, void *parent, void *child)
     }
   }
 
+  ngx_conf_merge_str_value(conf->cache_base_path,
+                           prev->cache_base_path,
+                           NULL);
+
   return NGX_CONF_OK;
 }
 
@@ -1451,7 +1460,17 @@ ngx_http_groonga_open_database_callback(ngx_http_groonga_loc_conf_t *location_co
     }
   }
 
-  location_conf->cache = grn_cache_open(context);
+  if (location_conf->cache_base_path.data &&
+      ngx_str_is_custom_path(&(location_conf->cache_base_path))) {
+    char cache_base_path[PATH_MAX];
+    grn_memcpy(cache_base_path,
+               location_conf->cache_base_path.data,
+               location_conf->cache_base_path.len);
+    cache_base_path[location_conf->cache_base_path.len] = '\0';
+    location_conf->cache = grn_persistent_cache_open(context, cache_base_path);
+  } else {
+    location_conf->cache = grn_cache_open(context);
+  }
   if (!location_conf->cache) {
     ngx_log_error(NGX_LOG_EMERG, data->log, 0,
                   "failed to open Groonga cache: %s",
@@ -1479,12 +1498,10 @@ ngx_http_groonga_close_database_callback(ngx_http_groonga_loc_conf_t *location_c
   ngx_http_groonga_context_init_query_logger(location_conf,
                                              data->pool,
                                              data->log);
-  grn_cache_current_set(context, location_conf->cache);
 
   grn_obj_close(context, location_conf->database);
   ngx_http_groonga_context_log_error(data->log);
 
-  grn_cache_current_set(context, NULL);
   grn_cache_close(context, location_conf->cache);
 }
 
@@ -1522,8 +1539,8 @@ ngx_http_groonga_init_process(ngx_cycle_t *cycle)
 
   grn_set_segv_handler();
 
-  rc = grn_ctx_init(context, GRN_NO_FLAGS);
-  if (rc != GRN_SUCCESS) {
+  context = grn_ctx_open(GRN_NO_FLAGS);
+  if (!context) {
     return NGX_ERROR;
   }
 
@@ -1548,7 +1565,10 @@ ngx_http_groonga_exit_process(ngx_cycle_t *cycle)
                                  ngx_http_groonga_close_database_callback,
                                  &data);
 
-  grn_ctx_fin(context);
+  if (context) {
+    grn_ctx_close(context);
+    context = NULL;
+  }
 
   grn_fin();
 
@@ -1620,6 +1640,13 @@ static ngx_command_t ngx_http_groonga_commands[] = {
     offsetof(ngx_http_groonga_loc_conf_t, default_request_timeout_msec),
     NULL },
 
+  { ngx_string("groonga_cache_base_path"),
+    NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF|NGX_CONF_TAKE1,
+    ngx_conf_set_str_slot,
+    NGX_HTTP_LOC_CONF_OFFSET,
+    offsetof(ngx_http_groonga_loc_conf_t, cache_base_path),
+    NULL },
+
   ngx_null_command
 };
 
